﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.ConditionActionEncoding {
  [Item("PredictionArrayCalculator", "")]
  [StorableClass]
  public class PredictionArrayCalculator : SingleSuccessorOperator {

    public ILookupParameter<IClassifierComparer> ClassifierComparerParameter {
      get { return (ILookupParameter<IClassifierComparer>)Parameters["ClassifierComparer"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> FitnessParameter {
      get { return (ILookupParameter<ItemArray<DoubleValue>>)Parameters["Fitness"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> PredictionParameter {
      get { return (ILookupParameter<ItemArray<DoubleValue>>)Parameters["Prediction"]; }
    }
    public IValueParameter<IItemDictionary<IAction, DoubleValue>> PredictionArrayParameter {
      get { return (IValueParameter<IItemDictionary<IAction, DoubleValue>>)Parameters["PredictionArray"]; }
    }
    public ILookupParameter<ItemArray<IClassifier>> MatchParameter {
      get { return (ILookupParameter<ItemArray<IClassifier>>)Parameters["MatchParameter"]; }
    }

    [StorableConstructor]
    protected PredictionArrayCalculator(bool deserializing) : base(deserializing) { }
    protected PredictionArrayCalculator(PredictionArrayCalculator original, Cloner cloner)
      : base(original, cloner) {
    }
    public PredictionArrayCalculator()
      : base() {
      Parameters.Add(new LookupParameter<IClassifierComparer>("ClassifierComparer"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Fitness"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Prediction"));
      Parameters.Add(new ValueLookupParameter<IItemDictionary<IAction, DoubleValue>>("PredictionArray"));
      Parameters.Add(new ScopeParameter("CurrentScope", "The current scope from which sub-scopes should be selected."));
      Parameters.Add(new ScopeTreeLookupParameter<IClassifier>("MatchParameter", "The matching encoding contained in each sub-scope which is used for selection."));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PredictionArrayCalculator(this, cloner);
    }

    public sealed override IOperation Apply() {
      IItemDictionary<IAction, DoubleValue> predictionArray = new ItemDictionary<IAction, DoubleValue>(ClassifierComparerParameter.ActualValue);
      IDictionary<IAction, double> fitnessSumPerAction = new Dictionary<IAction, double>(ClassifierComparerParameter.ActualValue);
      ItemArray<DoubleValue> fitnesses = FitnessParameter.ActualValue;
      ItemArray<DoubleValue> predictions = PredictionParameter.ActualValue;
      for (int i = 0; i < MatchParameter.ActualValue.Length; i++) {
        var action = MatchParameter.ActualValue[i].Action;
        if (predictionArray.ContainsKey(action)) {
          predictionArray[action].Value += predictions[i].Value * fitnesses[i].Value;
          fitnessSumPerAction[action] += fitnesses[i].Value;
        } else {
          predictionArray[action] = new DoubleValue(predictions[i].Value * fitnesses[i].Value);
          fitnessSumPerAction[action] = fitnesses[i].Value;
        }
      }

      foreach (var action in predictionArray.Keys) {
        if (fitnessSumPerAction[action] != 0) {
          predictionArray[action].Value /= fitnessSumPerAction[action];
        }
      }

      PredictionArrayParameter.ActualValue = predictionArray;

      return base.Apply();
    }
  }
}
